import copy
import datetime
import decimal
import enum
import json
import re
import uuid

# JSON DUMPS


class JsonLoadException(Exception):
    pass


def convert_to_dumps(data):
    if data == None:
        return None

    data_copy = copy.copy(data)

    if isinstance(data_copy, datetime.datetime):
        return data_copy.strftime('%Y-%m-%dT%H:%M:%S')
    elif isinstance(data_copy, datetime.date):
        return data_copy.strftime('%Y-%m-%d')
    elif isinstance(data_copy, uuid.UUID):
        return str(data_copy)
    elif isinstance(data_copy, decimal.Decimal):
        return str(data_copy)
    elif isinstance(data_copy, dict):
        for key in data_copy.keys():
            data_copy[key] = convert_to_dumps(data_copy[key])
        return data_copy
    elif isinstance(data_copy, list):
        for idx in range(0, len(data_copy)):
            data_copy[idx] = convert_to_dumps(data_copy[idx])

        return data_copy
    elif isinstance(data_copy.__class__, enum.EnumMeta):
        return data_copy.value
    elif isinstance(data_copy, str) or isinstance(data_copy, int) or isinstance(data_copy, float) or isinstance(data_copy, bool):
        return data_copy
    elif isinstance(data_copy, object):
        to_dict_method = getattr(data_copy, 'to_dict', None)
        if to_dict_method is not None and callable(to_dict_method):
            dict_attrs = data_copy.to_dict()
        else:
            attrs_fields = [k for k in data_copy.__dict__ if not callable(
                getattr(data_copy, k, None))]
            dict_attrs = {k: data_copy.__dict__[k] for k in attrs_fields}

        return convert_to_dumps(dict_attrs)
    else:
        return data_copy


def json_dumps(data, ensure_ascii=True, convert_before_dump=True):
    """
    Retorna a representação em json (string) do objeto recebido no parâmetro "data".

    É importante destacar que este método está preparado para as seguintes transformações:
    - datetime.datetime => '%Y-%m-%dT%H:%M:%S'
    - datetime.date => '%Y-%m-%d'
    - uuid.UUID => str(uuid.UUID)
    - Decimal => float

    Além disso, objetos (que não sejam de tipos primitivos, como str, float e bool), são
    tratados como dicionários, tendo todos os seus atributos considerados como chaves do json.

    Adicionalmente, se um objeto implementar um método "to_dict", a representação, em dicionário,
    desse objeto será obtida por meio da invocação deste método, antes da transformação do mesmo
    em json (permitindo customizar o modo como um objeto é serializado em json).
    """
    if convert_before_dump:
        data_copy = convert_to_dumps(data)
    else:
        data_copy = data

    return json.dumps(data_copy, ensure_ascii=ensure_ascii)


# JSON LOADS
def _loads_datetime_uuid(value):
    if not isinstance(value, str):
        return value

    matcher_datetime = re.compile(
        '^(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)$')
    matcher_date = re.compile('^(\d\d\d\d)-(\d\d)-(\d\d)$')
    matcher_uuid = re.compile(
        '^[A-Fa-f0-9]{8}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{12}$')

    match_datetime = matcher_datetime.search(value)
    match_date = matcher_date.search(value)
    match_uuid = matcher_uuid.search(value)

    if match_datetime:
        ano = int(match_datetime.group(1))
        mes = int(match_datetime.group(2))
        dia = int(match_datetime.group(3))
        hora = int(match_datetime.group(4))
        minuto = int(match_datetime.group(5))
        segundo = int(match_datetime.group(6))

        return datetime.datetime(year=ano, month=mes, day=dia, hour=hora, minute=minuto, second=segundo)
    elif match_date:
        ano = int(match_date.group(1))
        mes = int(match_date.group(2))
        dia = int(match_date.group(3))

        return datetime.date(year=ano, month=mes, day=dia)
    elif match_uuid:
        return uuid.UUID(value)
    else:
        return value


def _internal_loads(data):
    if isinstance(data, dict):
        for key in data.keys():
            data[key] = _internal_loads(data[key])
        return data

    elif isinstance(data, list):
        vector = []
        for item in data:
            vector.append(_internal_loads(item))
        return vector

    else:
        return _loads_datetime_uuid(data)


def _loads_to_class(load_data, model_class=None):
    if isinstance(load_data, list):
        return [_loads_to_class(item, model_class) for item in load_data]
    elif isinstance(load_data, dict):
        obj = model_class()
        for k in load_data:
            if hasattr(obj, k):
                setattr(obj, k, load_data[k])
        return obj
    else:
        return model_class()


def json_loads(str_json: str, model_class=None):
    """
    Interpreta a string json recebida no parâmetro "str_json", retornando:
    - Um dicionário ou uma lista de dicionários (se o parâmetro "model_class" for nulo)
    - Um objeto do tipo model_class, ou uma lista desses objetos, atribuido o valor
    das chaves correspondentes no json, para cada atributo com mesmo nome, no objeto.

    É importante destacar que este método está preparado para as seguintes transformações:
    - '%Y-%m-%dT%H:%M:%S' => datetime.datetime
    - '%Y-%m-%d' => datetime.date
    """
    try:
        if isinstance(str_json, str):
            data = json.loads(str_json)
        else:
            data = str_json

        load_data = _internal_loads(data)

        if model_class is None or (not isinstance(load_data, dict) and not isinstance(load_data, list)):
            return load_data
        else:
            return _loads_to_class(load_data, model_class)
    except Exception as e:
        msg = f"Erro interpretando json. Mensagem original do erro: {e}."
        msg += '\nCorpo do json recebido:\n'
        msg += f"{str_json}"

        raise JsonLoadException(msg)



from typing import List
from django.db import connections, transaction, connection
from sqlparams import SQLParams
from enum import Enum


class AbstractRepository:
    def __init__(self):
        self.alias_banco = "default"
        self.__con = connections[self.alias_banco]

    @staticmethod
    def _fetchall(cursor):
        desc = cursor.description
        return [
            dict(zip([col[0] for col in desc], row))
            for row in cursor.fetchall()
        ]

    @staticmethod
    def _fetchone(cursor):
        if cursor.rowcount == 0:
            return None
        desc = cursor.description
        return dict(zip([col[0] for col in desc], cursor.fetchone()))

    def set_connection(self, connection):
        self.__con = connection

    def begin(self):
        transaction.set_autocommit(False, using=self.alias_banco)

    def commit(self):
        transaction.commit(using=self.alias_banco)
        transaction.set_autocommit(True, using=self.alias_banco)

    def rollback(self):
        transaction.rollback(using=self.alias_banco)
        transaction.set_autocommit(True, using=self.alias_banco)

    def execute(self, sql: str, params: dict):
        with self.__con.cursor() as cursor:
            self.__execute_retornando_cursor(sql, params, cursor)
            return cursor.rowcount

    def fetchAll(self, sql: str, params: dict) -> List[dict]:
        with self.__con.cursor() as cursor:
            self.__execute_retornando_cursor(sql, params, cursor)
            retorno = AbstractRepository._fetchall(cursor)
        return retorno

    def fetchOne(self, sql: str, params: dict) -> List[dict]:
        with self.__con.cursor() as cursor:
            self.__execute_retornando_cursor(sql, params, cursor)
            retorno = AbstractRepository._fetchone(cursor)
        return retorno

    def __execute_retornando_cursor(self, sql: str, params: dict, cursor):
        if not isinstance(params, dict):
            params = params.__dict__
        sql2, params2 = SQLParams('named', 'format').format(sql, params)
        params2 = [elem.value if isinstance(
            elem, Enum) else elem for elem in params2]
        cursor.execute(sql2, params2)
        return cursor



from nsj_gcf_utils  .exception import ERPException
from nsj_gcf_utils.json_util import json_dumps, json_loads
from pydantic import ValidationError
from typing import Dict, List, Union, Tuple


def _format_tuple_error(error: Tuple[str, str]):
    return {
        'code': error[0],
        'message': error[1]
    }


def _format_erpexception_error(error: ERPException):
    return {
        'code': error.mope_code,
        'message': error.message
    }


def _format_list_error(error: Union[List[Tuple[str, str]], List[ERPException]]):
    result = []
    for e in error:
        if isinstance(e, tuple):
            formated = _format_tuple_error(e)
        elif isinstance(e, ERPException):
            formated = _format_erpexception_error(e)
        elif isinstance(e, str):
            e = (None, e)
            formated = _format_tuple_error(e)
        elif isinstance(e, Exception):
            e = (None, f'{e}')
            formated = _format_tuple_error(e)
        else:
            formated = _format_unknow_error()

        result.append(formated)

    return result


def _format_unknow_error():
    return {
        'code': None,
        'message': 'Erro desconhecido'
    }


def _format_pydantic_validation_error(error: ValidationError):
    result = []
    errors = json_loads(error.json())
    for e in errors:
        msg = f"Erro de validando campo '{e['loc'][0]}' de entrada. Mensagem do erro: {e['msg']}."
        e_tuple = (None, msg)
        result.append(_format_tuple_error(e_tuple))

    return result


def format_error_body(
    error: Union[
        Tuple[str, str],
        List[Tuple[str, str]],
        ERPException,
        List[ERPException],
        str,
        Exception,
        List[str],
        List[Exception],
        ValidationError
    ]
) -> List[Dict[str, str]]:

    if isinstance(error, tuple):
        return [_format_tuple_error(error)]
    elif isinstance(error, list):
        return _format_list_error(error)
    elif isinstance(error, ERPException):
        return [_format_erpexception_error(error)]
    elif isinstance(error, str):
        error = (None, error)
        return [_format_tuple_error(error)]
    elif isinstance(error, ValidationError):
        return _format_pydantic_validation_error(error)
    elif isinstance(error, Exception):
        error = (None, f'{error}')
        return [_format_tuple_error(error)]
    else:
        return [_format_unknow_error()]


def format_json_error(
    error: Union[
        Tuple[str, str],
        List[Tuple[str, str]],
        ERPException,
        List[ERPException],
        str,
        Exception,
        List[str],
        List[Exception],
        ValidationError
    ]
) -> List[Dict[str, str]]:
    return json_dumps(format_error_body(error))



def _convert_to(obj: object, new_dto_class):
    new_obj = new_dto_class()

    for attribute in new_obj.__dict__:
        if not hasattr(obj, attribute):
            continue

        attr_origem = getattr(obj, attribute, None)
        attr_destino = getattr(new_obj, attribute, None)

        # Pulando métodos
        if callable(attr_origem) or callable(attr_destino):
            continue

        setattr(new_obj, attribute, attr_origem)

    return new_obj


def convert_to(obj: object, new_dto_class):
    """
    Recebendo um objeto qualquer (parâmetro "obj"), e uma classe qualquer (parâmetro "new_dto_class"),
    este método cria uma instância da classe "new_dto_class", e copia todos os atributos do objeto
    "obj" para a nova instância, desde que haja um atributo com mesmo nome na classe de destino.

    Este método também está preparado para receber uma lista de objetos, e converter para uma nova
    lista de objetos do tipo "new_dto_class".
    """

    if isinstance(obj, list):
        return [_convert_to(item, new_dto_class) for item in obj]
    else:
        return _convert_to(obj, new_dto_class)



class ERPException(Exception):
    mope_code: str
    message: str

    def __init__(self, mope_code: str, message: str):
        super().__init__(f'{mope_code} - {message}')
        self.mope_code = mope_code
        self.message = message


class PaginationException(ERPException):
    def __init__(self, msg: str):
        super().__init__('0000-E001',
                         f'Erro nos parâmetros requisitados para paginação: {msg}')


class NotFoundException(Exception):
    pass
